# Imports
import numpy as np
import matplotlib.pyplot as plt

################################################################################

def mostrar_grid_search_results(rangoC, rangoGamma, resultados):
    '''
    Función auxiliar que muestra los parámetros del gridsearch.
    Entrada:
        rangoC: lista con los valores de C evaluados
        rangoGamma: lista con los valores de gamma evaluados
        resultados: arrglo de tamaño (len(rangoC),len(rangoGamma)) con los resultados
                    del grid-search
    '''
    fig, ax = plt.subplots(figsize=(6,6))
    im = ax.imshow(resultados, interpolation='nearest', cmap=plt.cm.Blues)
    ax.figure.colorbar(im, ax=ax)
    ax.set_ylabel('C')
    ax.set_ylim(-0.5,len(rangoC)-0.5)
    ax.set_xlim(-0.5,len(rangoGamma)-0.5)
    ax.set_yticks(range(len(rangoC)))
    ax.set_xticks(range(len(rangoGamma)))
    rangoC_labels = [str(c) for c in rangoC]
    ax.set_yticklabels(rangoC_labels)
    ax.set_xlabel('Gamma')
    rangoGamma_labels = [str(gamma) for gamma in rangoGamma]
    ax.set_xticklabels(rangoGamma_labels)
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                    rotation_mode="anchor")
    plt.title('Resultados del Grid Search')

################################################################################

def mostrar_matriz_confusion(cm, target_names,
                            normalize=False,
                            cmap=plt.cm.Blues):
    '''
    Muestra la matriz de confusión
    '''

    fig, ax = plt.subplots(figsize=(10,10))
    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)
    ax.figure.colorbar(im, ax=ax)
    ax.set(xticks=np.arange(cm.shape[1]),
            yticks=np.arange(cm.shape[0]),
            xticklabels=target_names, yticklabels=target_names,
            title= 'Matriz de confusion normalizada' if normalize else 'Matriz de confusion sin normalizar',
            ylabel='Etiqueta',
            xlabel='Predicción',
            xlim=(-0.5,9.5),
            ylim=(9.5,-0.5))

    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                rotation_mode="anchor")

    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(j, i, format(cm[i, j], fmt),
                    ha="center", va="center",
                    color="white" if cm[i, j] > thresh else "black")
    fig.tight_layout()

################################################################################

def graficar_distribucion(metadata, column):

    # Crear un gráfico de barras para la distribución de clases. El atributo columno es una lista de strings, cada columna se grafica como un subplot a la derecha.

    # Crear el gráfico
    fig, ax = plt.subplots(1, len(column), figsize=(10, 5))

    # Iterar sobre las columnas
    for i, col in enumerate(column):

        # Obtener la distribución de clases
        distribucion = metadata[col].value_counts()

        # Graficar la distribución de clases como un gráfico de barras horizontales
        ax[i].barh(distribucion.index, distribucion.values,  color='skyblue')

        # Agregar título y etiquetas
        ax[i].set_title(col)
        ax[i].set_ylabel('Clase')
        ax[i].set_xlabel('Cantidad de muestras')

        # Agregar grid
        ax[i].grid(axis='x', linestyle='--', alpha=0.6)

        # Agrego titulo al gráfico
        ax[i].set_title('Distribución por ' + col)
    
    # Ajustar el espacio entre los subplots
    fig.tight_layout()
    
    # Mostrar el gráfico
    plt.show()
